# TODO: remove innereye references
innereye_ssl_checkpoint = "hsharma_panda_explore_1638437076_357167ae"
innereye_ssl_checkpoint_binary = "PANDA_SimCLR_1654023538_7a0f1d00"
innereye_ssl_checkpoint_crck_4ws = "CRCK_SimCLR_1653673515_42d53d78"
innereye_ssl_checkpoint_crck_radiomics = "704b1af8-7c75-46ed-8460-d80a0e603194"

# outdated checkpoints
# innereye_ssl_checkpoint_crck_radiomics = updated_transforms:updated_transforms_1636471522_5473e3ff
